#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <sched.h>
#include <glib.h>
#include <signal.h>

#include <sys/types.h>
#include <sys/wait.h>

static int spawn_helper(void *data)
{
    char ** argv = (char **) data;
    int ret;
    g_return_val_if_fail (argv != NULL, -1);

    ret = execv(argv[0], argv);
    if (ret < 0)
        perror("execv");

    return 0;
}

static int system_helper(void *data)
{
    const char *cmd = (const char*)data;
    int ret;
    gchar **argv = NULL;

    g_return_val_if_fail (cmd != NULL, -1);

    if (!g_shell_parse_argv (cmd, NULL, &argv, NULL)) {
        return -1;
    }

    ret = execv(argv[0], argv);
    if (ret < 0)
        perror("execv");

    g_strfreev (argv);

    return 0;
}

int good_fork(int (*fn)(void *), void *arg)
{
    int pid;
    char st[128 * 1024] __attribute__ ((aligned));
    pid = clone(fn, st+sizeof(st), SIGCHLD, arg);
    if (pid < 0) {
        perror("Error clone");
        return -1;
    }
    return pid;
}

int good_system(const char *cmdline)
{
    int status;
    int pid;

    pid = good_fork(system_helper, (void*)cmdline);
    if (pid < 0) {
        return -1;
    }

    waitpid(pid, &status, 0);
    return status;
}

int good_spawn(char * const args[])
{
    int pid;

    pid = good_fork(spawn_helper, (void*)args);
    if (pid < 0) {
        return -1;
    }
    return pid;
}
